#include <stdio.h>
#include <stdlib.h>
#include <string.h>

//структура, що містить дані про книги
struct book {
  char author[50];
  char title[100];
  int year;
  int pages;
  float cost;
};

//------------------------------------------------------------------------------------------

//однозв'язний список
struct ListNode {
  struct book data;
  struct ListNode* next;
};

//------------------------------------------------------------------------------------------

//конвертація масиву структур з даними в одзнозв'язний список
void appendNode(struct ListNode** head, struct book data) {
  struct ListNode* newNode = (struct ListNode*)malloc(sizeof(struct ListNode));
  newNode->data = data;
  newNode->next = NULL;

  if (*head == NULL) {
    *head = newNode;
  } else {
    struct ListNode* last = *head;
    while (last->next != NULL) {
      last = last->next;
    }
    last->next = newNode;
  }
}

//------------------------------------------------------------------------------------------

//виведення даних у вигляді таблиці
void printList(struct ListNode* head) {
  printf("| %-6s | %-23s | %-30s | %-4s | %-6s | %-8s |\n", "Index", "Author", "Book name", "Year", "Page", "Price, $");
  printf("|--------|-------------------------|--------------------------------|------|--------|----------|\n");
  int index = 1;
  while (head != NULL) {
    printf("| %-6d | %-23s | %-30s | %-4d | %-6d | %-8.2f |\n",
    index, head->data.author, head->data.title, head->data.year, head->data.pages, head->data.cost);
    head = head->next;
    index++;
  }
}

//------------------------------------------------------------------------------------------

//звільнення пам'яті від однозв'язних списків
void freeList(struct ListNode* head) {
  struct ListNode* current = head;
  struct ListNode* next;

// Перевірка, чи не є голова списку рівною NULL
  if (head == NULL)
    return;

  while (current->next != NULL) {
      next = current->next;
      free(current);
      current = next;
  }
  free(current);
  head = NULL;
}


//------------------------------------------------------------------------------------------

//видалення книги за введеним індексом
void deleteBook(struct ListNode** head, int index) {
  if (index < 0) {
      printf("Invalid index. No book deleted.\n");
      return;
  }

  struct ListNode* current = *head;
  struct ListNode* prev = NULL;
  int i = 0;

  while (current != NULL && i < index) {
    prev = current;
    current = current->next;
    i++;
  }

  if (current == NULL) {
    printf("Invalid index. No book deleted.\n");
    return;
  }

  char response[5];
  int choice;
  while(1){
    printf("Are you sure you want to delete the book '%s' by %s?\n 1 - Yes\n 2 - No\n Enter your choice:  ", current->data.title, current->data.author);
    scanf(" %s", response);
    if(!CheckIfInt(response)){
      printf("Wrong format. Please enter '1' or '2'\n");
    } else{
      choice = atoi(response);
      break;
    }
  }
  if (choice != 1) {
    printf("Deletion cancelled.\n");
    return;
  }

  if (prev == NULL) {
    *head = current->next;
  } else {
    prev->next = current->next;
  }

  printf("Book '%s' by %s deleted successfully!\n", current->data.title, current->data.author);
  free(current);
}

//------------------------------------------------------------------------------------------

//функція для видалення книжок в котрих менше 90 сторінок та написані до 1980 року
void deleteBooksLessThan90PagesBefore1980(struct ListNode** head) {
  if (head == NULL || *head == NULL)
    return;

  struct ListNode* current = *head;
  struct ListNode* prev = NULL;
  int deletionConfirmed = 0;

  while (current != NULL && current->next != NULL) {
    if (current->data.year < 1980 && current->data.pages < 90) {
      deletionConfirmed = 1;
      break;
    }
    prev = current;
    current = current->next;
  }

  if (deletionConfirmed) {
    printf("The following books will be deleted:\n");
    current = *head;
    while (current != NULL && current->next != NULL) {
      if (current->data.year < 1980 && current->data.pages < 90) {
          printf("Title: %s, Author: %s\n", current->data.title, current->data.author);
      }
      current = current->next;
    }

    printf("Are you sure you want to delete all books with less than 90 pages and published before 1980?\n 1 - Yes\n 2 - No\n Enter your choice: ");
    char response[5];
    int choice;
    while(1) {
      scanf(" %s", response);
      if (!CheckIfInt(response)) {
          printf("Wrong format. Please enter '1' or '2'\n");
      } else {
          choice = atoi(response);
          break;
      }
    }
    if (choice != 1) {
      printf("Deletion cancelled.\n");
      return;
    }
  }

  current = *head;
  prev = NULL;
  while (current != NULL && current->next != NULL) {
    if (current->data.year < 1980 && current->data.pages < 90) {
      struct ListNode* temp = current;
      if (prev == NULL) {
          *head = current->next;
      } else {
          prev->next = current->next;
      }
      current = current->next;
      free(temp);
    } else {
      prev = current;
      current = current->next;
    }
  }
  printf("Books deleted successfully!\n");
}



//------------------------------------------------------------------------------------------

//збереження даних у файл
void writeBooksToFile(const char* filename, struct book books[], int num_books) {
  FILE* file = fopen(filename, "w");
  if (file == NULL) {
    printf("File opening failed.\n");
    return;
  }

  for (int i = 0; i < num_books; i++) {
    fprintf(file, "%s, %s, %d, %d, %.2f\n", books[i].author, books[i].title, books[i].year, books[i].pages, books[i].cost);
  }

  fclose(file);
}

//------------------------------------------------------------------------------------------

//зчитування даних з файлу
int readBooksFromFile(const char* filename, struct book books[], int max_books) {
  FILE* file = fopen(filename, "r");
  if (file == NULL) {
    printf("File opening failed.\n");
    return 1;
  }

  char line[256];
  int num_books = 0;

  while (fgets(line, sizeof(line), file) != NULL) {
    char author[50], title[100];
    int year, pages;
    float cost;

    sscanf(line, "%[^,], %[^,], %d, %d, %f", author, title, &year, &pages, &cost);

    strcpy(books[num_books].author, author);
    strcpy(books[num_books].title, title);
    books[num_books].year = year;
    books[num_books].pages = pages;
    books[num_books].cost = cost;

    num_books++;
    if (num_books >= max_books) {
      printf("Too many books.\n");
      break;
    }
  }

  fclose(file);
  return num_books;
}

//------------------------------------------------------------------------------------------

//сортування книжок за автором (в алфавітному порядку)
void sortByAuthor(struct ListNode** head) {
  if (*head == NULL || (*head)->next == NULL) {
    return;
  }

  struct ListNode* current = *head;
  struct ListNode* index = NULL;
  struct book temp;

  while (current != NULL) {
    index = current->next;

    while (index != NULL) {
      if (strcmp(current->data.author, index->data.author) > 0) {
        temp = current->data;
        current->data = index->data;
        index->data = temp;
      }
      index = index->next;
    }
    current = current->next;
  }
}

//------------------------------------------------------------------------------------------

//сортування книжок за назвою (в алфавітному порядку)
void sortByTitle(struct ListNode** head) {
  if (*head == NULL || (*head)->next == NULL) {
    return;
  }

  struct ListNode* current = *head;
  struct ListNode* index = NULL;
  struct book temp;

  while (current != NULL) {
    index = current->next;

    while (index != NULL) {
      if (strcmp(current->data.title, index->data.title) > 0) {
        temp = current->data;
        current->data = index->data;
        index->data = temp;
      }
      index = index->next;
    }
    current = current->next;
  }
}

//------------------------------------------------------------------------------------------

//сортування книжок за роком написання (від найстарішої до найновішої)
void sortByYear(struct ListNode** head) {
  if (*head == NULL || (*head)->next == NULL) {
    return;
  }

  struct ListNode* current = *head;
  struct ListNode* index = NULL;
  struct book temp;

  while (current != NULL) {
    index = current->next;

    while (index != NULL) {
      if (current->data.year > index->data.year) {
        temp = current->data;
        current->data = index->data;
        index->data = temp;
      }
      index = index->next;
    }
    current = current->next;
  }
}

//------------------------------------------------------------------------------------------

//сортування книжок за ціною (від найдешевших до найдорожчих)
void sortByCost(struct ListNode** head) {
  if (*head == NULL || (*head)->next == NULL) {
    return;
  }

  struct ListNode* current = *head;
  struct ListNode* index = NULL;
  struct book temp;

  while (current != NULL) {
    index = current->next;

    while (index != NULL) {
      if (current->data.cost > index->data.cost) {
        temp = current->data;
        current->data = index->data;
        index->data = temp;
      }
      index = index->next;
    }
    current = current->next;
  }
}

//------------------------------------------------------------------------------------------

//сортування за кількістю сторінок (від найкоротших до найдовших)
void sortByPages(struct ListNode** head) {
  if (*head == NULL || (*head)->next == NULL) {
    return;
  }

  struct ListNode* current = *head;
  struct ListNode* index = NULL;
  struct book temp;

  while (current != NULL) {
    index = current->next;

    while (index != NULL) {
      if (current->data.pages > index->data.pages) {
        temp = current->data;
        current->data = index->data;
        index->data = temp;
      }
      index = index->next;
    }
    current = current->next;
  }
}

//------------------------------------------------------------------------------------------

// Фунція для перевірки на введений формат даних
int CheckIfInt(const char *input) {
  int i = 0;
  while (input[i] != '\0') {
    if (!isdigit(input[i]) && input[i] != '-' && input[i] != '.') {
      return 0;// якщо не цифра
    }
    i++;
  }

  return 1;// якщо цифра
}





int main() {
  int iMaxBooks = 100, iNumberOfBooks = 0;
  struct book books[iMaxBooks];
  struct ListNode* head = NULL;
  char CFileName[255];
  char cOperationChoice[5];
  int iOperationChoice;
  FILE* file;
  int iOpened=0;
  char mess;

  do {
    printf("\nChoose operation:\n\
    1 - File\n\
    2 - Sort\n\
    3 - Add\n\
    4 - Delete\n\
    5 - Quit\n\
    Enter your choice: ");
    scanf(" %s", &cOperationChoice);
    if(CheckIfInt(cOperationChoice)){
      mess = getchar();
      iOperationChoice = atoi(cOperationChoice);

      switch(iOperationChoice) {
        case 1: {
          printf("1 - open\n\
          2 - new\n\
          3 - save\n\
          Enter your choice: ");
          int fileChoice;
          scanf("%d", &fileChoice);
          switch(fileChoice) {
            case 1: {
              printf("Enter the name of the file to open: ");
              scanf("%s", CFileName);
              file = fopen(CFileName, "r+");
              if (file == NULL) {
                printf("File opening failed. Make sure '%s' exists.\n", CFileName);
                break;
              } else iOpened=1;
              iNumberOfBooks = readBooksFromFile(CFileName, books, iMaxBooks);
              if (iNumberOfBooks == 0) {
                printf("No books read from the file.\n");
              }
              printf("| %-6s | %-23s | %-30s | %-4s | %-6s | %-8s |\n", "Index", "Author", "Book name", "Year", "Page", "Price, $");
              printf("|--------|-------------------------|--------------------------------|------|--------|----------|\n");
              for (int i = 0; i < iNumberOfBooks; i++) {
                printf("| %-6d | %-23s | %-30s | %-4d | %-6d | %-8.2f |\n",
                i+1, books[i].author, books[i].title, books[i].year, books[i].pages, books[i].cost);
              }

              for (int i = 0; i < iNumberOfBooks; i++) {
                appendNode(&head, books[i]);
              }
              break;
              }
            case 2: {
              char fileName[100]; // Максимальна довжина назви файлу
              printf("Enter the name of the new file: ");
              scanf("%s", fileName);
              FILE* newFile;
              newFile = fopen(fileName, "w");
              if (newFile == NULL) {
                printf("File creation failed.\n");
                break;
              } else {
                printf("File created successfully!");
              }
              fclose(newFile);
              break;
              }
            case 3: {
              char fileName[100]; // Максимальна довжина назви файлу
              printf("Enter the name of the file to save: ");
              scanf("%s", fileName);
              writeBooksToFile(fileName, books, iNumberOfBooks);
              break;
            }
            default:
              printf("Wrong option. Please enter '1', '2' or '3'.\n");
            }
              break;
          }
        case 2: {
          printf("1 - by author\n\
          2 - by book title\n\
          3 - by year\n\
          4 - by pages\n\
          5 - by price\n");
          int sortChoice;
          printf("Enter your choice: ");
          scanf(" %d", &sortChoice);
          switch(sortChoice) {
            case 1:
              sortByAuthor(&head);
              break;
            case 2:
              sortByTitle(&head);
              break;
            case 3:
              sortByYear(&head);
              break;
            case 4:
              sortByPages(&head);
              break;
            case 5:
              sortByCost(&head);
              break;
            default:
              printf("Wrong option. Please enter '1', '2', '3', '4', or '5'.\n");
          }
          printf("\nBooks in the linked list after operation:\n");
          printList(head);
          break;
        }
        case 3: {
          struct book newBook;
          char temp[6];

          printf("Enter author: ");
          fgets(newBook.author, 50, stdin );
          newBook.author[ strcspn( newBook.author, "\n" ) ] = '\0';

          printf("Enter title: ");
          fgets(newBook.title, 50, stdin );
          newBook.title[ strcspn( newBook.title, "\n" ) ] = '\0';

          printf("Enter year: ");
          scanf("%s", temp);

          if (CheckIfInt(temp)) {
              int year = atoi(temp);
              if (year >= 1 && year <= 2024) {
                newBook.year = year;
              } else {
                printf("Wrong format, it has to be a number from 1 to 2024\n");
                break;
              }
          } else {
              printf("Wrong format, it has to be a number between 0 and 2024\n");
          }


          printf("Enter pages: ");
          scanf("%s", temp);

          if (CheckIfInt(temp)) {
              int pages = atoi(temp);
              if (pages > 0) {
                newBook.pages = pages;
              } else {
                printf("Number of pages must be greater than 0\n");
                break;
              }
          } else {
              printf("Wrong format, it has to be number greater than 0\n");
          }

          printf("Enter cost: ");
          scanf("%s", temp);

          if (CheckIfInt(temp)) {
              float cost = atof(temp);
              if (cost > 0) {
                newBook.cost = cost;
              } else {
                printf("Cost must be greater than 0\n");
                break;
              }
          } else {
              printf("Wrong format, it has to be number greater than 0\n");
          }

          appendNode(&head, newBook);

          printf("\nBook added successfully!\n");

          printf("\nBooks in the linked list after operation:\n");
          printList(head);
          break;
        }
        case 4: {
          int deleteChoice;
          printf("1. Delete book\n\
          2. Delete books with less than 90 pages and written before 1980 year\n");
          scanf("%d", &deleteChoice);

          switch (deleteChoice) {
            case 1: {
              int index;
              printf("Enter the index of the book to delete: ");
              scanf("%d", &index);

              deleteBook(&head, index - 1);
              break;
            }
            case 2: {
              deleteBooksLessThan90PagesBefore1980(&head);
              break;
            }
            default:
              printf("Wrong option. Please enter '1' or '2'.\n");
              break;
          }

          printf("\nBooks in the linked list after operation:\n");
          printList(head);
          break;
        }
        case 5:
          break;
        default:
          printf("\nInvalid choice. Please enter '1', '2', '3', '4', or '5'.\n");
      }
    } else {
        printf("Wrong input. Please enter '1', '2', '3', '4', or '5'.");
    }

  } while (iOperationChoice != 5);

  freeList(head);

  if(iOpened)
    fclose(file);

  return 0;
}
